#pragma once

#include <string>
#include <stdexcept>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>

// 
//
class DataExportCommsBase
{
public:
    virtual ~DataExportCommsBase() { }

    virtual void write(const void* data, size_t length) = 0;
    virtual void read(void* outData, size_t length) = 0;

    template<typename T>
    void readValue(T& outValue)
    {
        read(&outValue, sizeof(T));
    }

    template<typename T>
    void writeValue(const T& value)
    {
        write(&value, sizeof(T));
    }

    void writeString(const std::string& str);
    void readString(std::string& outStr);
};

//
//
class NamedPipe : public DataExportCommsBase
{
public:

    NamedPipe(const std::string& pipeName, bool isOwner);
    ~NamedPipe();

    virtual void write(const void* data, size_t length);
    virtual void read(void* outData, size_t length);    

private:
    int         mFileDescriptor;
    bool        mIsOwner;
    std::string mPipeName;
};


class UnixSocketClosedException : public std::runtime_error
{
public:
    UnixSocketClosedException() : std::runtime_error("socket closed") { }
};


class VirtualSocketPath : public std::string
{
public:
    VirtualSocketPath(const std::string& str)
    {
        static_cast<std::string&>(*this) = std::string(" ") + str;
        (*this)[0] = '\0'; 
    }
};

class UnixSocket : public DataExportCommsBase
{
public:
    enum Mode
    {
        Client,
        Server
    };

    UnixSocket(const std::string& path, Mode mode)
    {
        mFd = -1;

        if (path.length() == 0)
            throw std::runtime_error("Invalid socket path");

        // Create socket
        if ((mFd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
            throw std::runtime_error("Failed to create socket");
            
        // Set identity/socket name
        // Path is a file path that will be created in the filesystem, 
        // or prefix with /0 to create a virtual socket identifier that 
        // won't be created as a file
        struct sockaddr_un addr;
        memset(&addr, 0, sizeof(addr));
        addr.sun_family = AF_UNIX;
        if (path[0] == '\0') 
        {
            *addr.sun_path = '\0';
            strncpy(addr.sun_path+1, path.c_str()+1, sizeof(addr.sun_path)-2);
        } 
        else 
        {
            strncpy(addr.sun_path, path.c_str(), sizeof(addr.sun_path)-1);
            unlink(path.c_str());
        }

        if (mode == Server)
        {           
            // bind 
            if (bind(mFd, (struct sockaddr*)&addr, sizeof(addr)) == -1) 
            {                
                throw std::runtime_error("Failed to bind server socket");
            }

            // listen
            if (listen(mFd, 5) == -1) 
            {
                throw std::runtime_error("Failed to start listening for connections");
            }
        }      
        else 
        {
            // client mode
            if (connect(mFd, (struct sockaddr*)&addr, sizeof(addr)) == -1) 
            {
                throw std::runtime_error("Failed to connect to server");
            }
        }          
    }


    UnixSocket(int fd)
        : mFd(fd)
    {
    }

    ~UnixSocket()
    {
        if (mFd > 0)
            ::close(mFd);
    }

    std::unique_ptr<UnixSocket> acceptNewConnection()
    {
        int cl;
        if ( (cl = accept(mFd, NULL, NULL)) == -1) 
        {
            throw std::runtime_error("Socket accept error");
        }

        return std::unique_ptr<UnixSocket>(new UnixSocket(cl));
    }

    void write(const void* data, size_t length)
    {
        size_t remaining = length;
        while (remaining > 0)
        {
            int numWritten = ::write(mFd, data, remaining);
            if (numWritten == -1)
                throw std::runtime_error("socket write error");
            else if (numWritten == 0)
                throw UnixSocketClosedException();
            else 
            {
                remaining -= (size_t)numWritten;
                data = (const char*)(data) + numWritten;
            }
        }
    }

    void read(void* outData, size_t length) 
    {
        // TODO: handle different write lengths
        int numRead = ::read(mFd, outData, length);
        if (numRead == 0)
            throw UnixSocketClosedException();
        else if (numRead == -1)
            throw std::runtime_error("socket read error");                        
    }
private:
    int mFd;
};
