/*
 * Copyright (C) 1996-2016 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

#include "squid.h"
#include "helpers/defines.h"
#include "ssl/certificate_db.h"
#include "ssl/crtd_message.h"

#include <cstring>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <string>
#if HAVE_GETOPT_H
#include <getopt.h>
#endif

// added by AG
#include <fstream>
#include <time.h>
#include <sstream>
#include <unistd.h>
#include <vector>

namespace kantar
{
    //
    // Build number

    int gBuildNumber = 9;

    // 
    // Utils

    template<typename T>
    std::string toString(const T& value)
    {
        std::stringstream format;
        format << value;
        return format.str();
    }

    std::string readAllText(const std::string& path)
    {
        std::basic_ifstream<char> file(path.c_str());
        if (!file.is_open())
            throw std::runtime_error(std::string("Failed to open file '") + path + "'" + " reason: " + std::strerror(errno));

        std::string fileContents(
                (std::istreambuf_iterator<char>(file)),
                std::istreambuf_iterator<char>());

        return fileContents;
    }


    //
    // Logging

    class Log 
    {
    public:
        virtual ~Log() { }
        virtual void writeLine(const std::string& str) 
        {            
            time_t _tm = time(NULL);
            struct tm* curtime = localtime (&_tm);
            std::stringstream format;
            char dateStr[512];
            strcpy(dateStr, asctime(curtime));
            dateStr[strlen(dateStr) - 1] = 0x0; // put null terminator one space before to chop off newline that asctime puts at the end 

            format << dateStr << "\t" << str << "\n";
            writeImpl(format.str());
        
        } 
    private:
        virtual void writeImpl(const std::string& str) = 0;  
    };

    class FileLog : public Log
    {
    public:
        virtual void writeImpl(const std::string& str)
        {
            std::ofstream file("/usr/var/logs/ssl_crtd.log", std::ofstream::out | std::ofstream::app);
            file << str;
        }
    };


    class NullLog : public Log
    {
    public:
        virtual void writeImpl(const std::string& str)
        {
            // don't do anything!
        }
    };
    
    // Log instance    
    Log* log;


    //
    // String util_sccsid
    
    bool hasEnding (const std::string& fullString, const std::string& ending) 
    {
        if (fullString.length() >= ending.length())         
            return (0 == fullString.compare (fullString.length() - ending.length(), ending.length(), ending));        
        else         
            return false;        
    }

    std::string replaceAll(const std::string& str, const std::string& sourceText, const std::string& replaceWithText)
    {
        std::string strCopy(str);
        std::string::size_type foundPos;
        while ((foundPos = strCopy.find(sourceText)) != std::string::npos)
        {
            strCopy.replace(foundPos, sourceText.length(), replaceWithText);
        }
        return strCopy;
    }


    // 
    // Decrypt certificates

    

    //
    // Certificate loading and sending

    class Certificate
    {
        public:
            std::string certChainWithPrivKey;        
    };

    class Mapping
    {
    public:
        std::string filter;
        std::string certName;
    };

    namespace paths
    {        
        const std::string kConfigPath = "/tmp/ssl_decrypt/";
        const std::string kMappingConfigPath = paths::kConfigPath + "/mapping.conf";
        const std::string kEnableLoggingMarkerFilePath = "/home/tdd/focalmeter/config/enable_ssl_logging";
    }
        

    void loadNamedCertificate(const std::string& name, Certificate& outCert)
    {        
        std::stringstream bufferToWrite;
        const std::string certPath = paths::kConfigPath + name + ".cert";
        const std::string keyPath = paths::kConfigPath + name + ".key";
        const std::string interCertPath = paths::kConfigPath + name + ".intermediate.cert";

        // Log for debug purpose
        log->writeLine("###### Loading certificate for domain: " + name + " ######");

        // Read certificate and private key from provided files
        Ssl::X509_Pointer cert;
        Ssl::EVP_PKEY_Pointer pkey;
        Ssl::readCertAndPrivateKeyFromFiles(cert, pkey, certPath.c_str(), keyPath.c_str());

        // Check that certificates and private key are loaded successfully
        if (!cert)    
            throw std::runtime_error("Failed to load SSL certificate");

        if (!pkey)
            throw std::runtime_error("Failed to load SSL private key");        

        log->writeLine(std::string("cert name: ") + cert.get()->name);

        // Write the certificate and private key to output
        std::string certBuffer;
        if (!Ssl::writeCertAndPrivateKeyToMemory(cert, pkey, certBuffer))
            throw std::runtime_error("Cannot write ssl certificate or/and private key to memory.");

        bufferToWrite << certBuffer.length() << std::endl;
        bufferToWrite.write(certBuffer.c_str(), certBuffer.length());

        // Check if intermediate certificate file exists
        bool hasIntermediate;
        {            
            std::ifstream file(interCertPath.c_str());
            hasIntermediate = file.is_open();
        }

        // Load intermediate certificates if needed
        if (hasIntermediate)
        {
            Ssl::X509_STACK_Pointer intermediateCert;

            // allocate new interemediate certificate chain
            intermediateCert.reset(sk_X509_new_null());
            if (!intermediateCert)
                throw std::runtime_error(std::string("Failed to allocate SSL intermediate certificate: ") + interCertPath);

            // Read intermediate certificate file
            if(0 != Ssl::readSslX509CertificateChain(interCertPath.c_str(), intermediateCert.get()))
                throw std::runtime_error(std::string("Failed to load SSL intermediate certificate: ") + interCertPath);

            // write intermediate certificate to output
            int num_certs = sk_X509_num(intermediateCert.get());
            if(num_certs > 0)
            {
                log->writeLine("write number of intermediate file to output (num_certs: " + std::to_string(num_certs) + ")");
                bufferToWrite << num_certs << std::endl;

                // For all certificates in the cahin
                for(int ii=0; ii<num_certs; ++ii)
                {
                    log->writeLine("Reading intermediate cert");

                    // Get the certificate from the chain
                    X509* intermediate_cert = sk_X509_value(intermediateCert.get(), ii);

                    // Increment reference counter as Ssl::appendCertToMemory is taking an X509_Pointer as input parameter
                    // and X509_Pointer will automatically call X509_free when the method is returning
                    CRYPTO_add(&intermediate_cert->references, 1, CRYPTO_LOCK_X509);

                    // Log certificate name for debug
                    log->writeLine(std::string("intermediate cert name: ") + intermediate_cert->name);

                    // Write certificate to output string buffer
                    std::string buffer;
                    if (!Ssl::appendCertToMemory(intermediate_cert, buffer))
                    {
                        log->writeLine("Cannot write ssl intermediate certificate to memory");
                        throw std::runtime_error("Cannot write ssl intermediate certificate to memory");
                    }

                    log->writeLine("wrote intermediate cert to memory buffer");

                    // append intermediate certificate to output
                    bufferToWrite << buffer.length() << std::endl;
                    bufferToWrite.write(buffer.c_str(), buffer.length());

                    log->writeLine("appended intermediate cert to output buffer chain");
                }
                log->writeLine("no more intermediate certificates");
                bufferToWrite << "0\n";
            } else {
                log->writeLine("No intermediate certificate to write");
                bufferToWrite << "0\n";
            }            
        }

        outCert.certChainWithPrivKey = bufferToWrite.str();
        log->writeLine("###### Certificate for domain: " + name + " loaded successfully ######");
    } 
  
    // Load certificates and config from filessystem
    std::map<std::string, Certificate*> gCertificates; // key is name of certificate
    std::vector<Mapping> gMappings;

    void loadCertificateConfig()
    {
        std::ifstream configFile(paths::kMappingConfigPath.c_str());
        std::string line;
        while (!configFile.eof())
        {
            std::getline(configFile, line);

            if (line.length() == 0)
                continue;
            std::istringstream parser(line);

            std::string filter, certName, validFrom;            
            parser >> filter >> certName >> validFrom;

            if (parser.fail())
                std::runtime_error("error while parsing certificate mapping file");

            if (gCertificates.count(certName) == 0)
            {
                // need to load cert
                Certificate* cert = new Certificate();
                loadNamedCertificate(certName, *cert);
                gCertificates[certName] = cert;

                //std::cout << "loaded cert " << certName << std::endl;
            }

            Mapping newMapping;
            newMapping.filter = filter;
            newMapping.certName = certName;
            gMappings.push_back(newMapping);
        }
    }

    const Certificate* getCertificateForDomain(const std::string& domain)
    {              
        log->writeLine(std::string("Looking for certificate... ") + domain);

        std::string cleanedDomain = replaceAll(domain, "*", "");
        log->writeLine("cleaned domain: " + cleanedDomain);

        for (std::vector<Mapping>::const_iterator it = gMappings.begin();
             it != gMappings.end();
             ++it)
        {
            log->writeLine(std::string("comparing ") + cleanedDomain + " and filter " + it->filter);

            // try to compare the requested hostname and filter, 
            // testing to see if one is an ending of the other.
            // The reason for trying both ways round is that the requested
            // domain can be a general wildcard domain name, such as *.example.com
            // instead of the full requested website name.  The filter string may
            // be more specific, e.g. specific.example.com, we'd still like this 
            // certificate to match the domain

            if (hasEnding(cleanedDomain, it->filter) || 
                hasEnding(it->filter, cleanedDomain))
            {
                log->writeLine("** match");
                return gCertificates.at(it->certName);                
            }
            else 
            {
                log->writeLine("** NO match");
            }
        }  

        throw std::runtime_error("Failed to find matching certificate");
    }


    void sendCertificateForDomain(const std::string& domain)
    {
        log->writeLine(std::string("creating certificate for '") + domain + "'");

        // Load certificate from files                    
        const Certificate* certificate = getCertificateForDomain(domain);

        log->writeLine("Completed call to 'getCertificateForDomain'");
        if (certificate)
            log->writeLine("Got a certificate object");
        else 
            log->writeLine("ERROR: Certificate object is null!");

        log->writeLine("About to write certificate to output");

        Ssl::CrtdMessage response_message(Ssl::CrtdMessage::REPLY);
        response_message.setCode("OK");
        response_message.setBody(certificate->certChainWithPrivKey);

        // Use the '\1' char as end-of-message character
        std::cout << response_message.compose() << '\1' << std::flush;

        log->writeLine(">>>>>>>>>>>>>>> SENT response");
    }
}


/** Changes by AG:
  *
  * - Replaced processNewRequest to a stored SSL certificate for supported domains,
  *   instead of generating a dynamic one with Squid's own Certificate Authority certificate
  * - Removed code that generates or checks for the presence of a certificate database 
  */

/**
 \defgroup ssl_crtd ssl_crtd
 \ingroup ExternalPrograms
 \par
    Because the standard generation of ssl certificate for
    sslBump feature, Squid must use external process to
    actually make these calls. This process generate new ssl
    certificates and worked with ssl certificates disk cache.
    Typically there will be five ssl_crtd processes spawned
    from Squid. Communication occurs via TCP sockets bound
    to the loopback interface. The class in helper.h are
    primally concerned with starting and stopping the ssl_crtd.
    Reading and writing to and from the ssl_crtd occurs in the
    \link IPCacheAPI IP\endlink and the dnsservers occurs in
    the \link IPCacheAPI IP\endlink and \link FQDNCacheAPI
    FQDN\endlink cache modules.

 \section ssl_crtdInterface Command Line Interface
 \verbatim
usage: ssl_crtd -hv -s ssl_storage_path -M storage_max_size
    -h                   Help
    -v                   Version
    -s ssl_storage_path  Path to specific disk storage of ssl server
                         certificates.
    -M storage_max_size  max size of ssl certificates storage.
    -b fs_block_size     File system block size in bytes. Need for processing
                         natural size of certificate on disk. Default value is
                         2048 bytes."

    After running write requests in the next format:
    <request code><whitespace><body_len><whitespace><body>
    There are two kind of request now:
    new_certificate 14 host=host.dom
        Create new private key and selfsigned certificate for "host.dom".

    new_certificate xxx host=host.dom
    -----BEGIN CERTIFICATE-----
    ...
    -----END CERTIFICATE-----
    -----BEGIN RSA PRIVATE KEY-----
    ...
    -----END RSA PRIVATE KEY-----
        Create new private key and certificate request for "host.dom".
        Sign new request by received certificate and private key.

usage: ssl_crtd -c -s ssl_store_path\n
    -c                   Init ssl db directories and exit.

 \endverbatim
 */

static const char *const B_KBYTES_STR = "KB";
static const char *const B_MBYTES_STR = "MB";
static const char *const B_GBYTES_STR = "GB";
static const char *const B_BYTES_STR = "B";

/**
  \ingroup ssl_crtd
 * Get current time.
*/
time_t getCurrentTime(void)
{
    struct timeval current_time;
#if GETTIMEOFDAY_NO_TZP
    gettimeofday(&current_time);
#else
    gettimeofday(&current_time, NULL);
#endif
    return current_time.tv_sec;
}

/**
  \ingroup ssl_crtd
 * Parse bytes unit. It would be one of the next value: MB, GB, KB or B.
 * This function is caseinsensitive.
 */
/* removed by AG, not referenced anymore
static size_t parseBytesUnits(const char * unit)
{
    if (!strncasecmp(unit, B_BYTES_STR, strlen(B_BYTES_STR)) ||
            !strncasecmp(unit, "", strlen(unit)))
        return 1;

    if (!strncasecmp(unit, B_KBYTES_STR, strlen(B_KBYTES_STR)))
        return 1 << 10;

    if (!strncasecmp(unit, B_MBYTES_STR, strlen(B_MBYTES_STR)))
        return 1 << 20;

    if (!strncasecmp(unit, B_GBYTES_STR, strlen(B_GBYTES_STR)))
        return 1 << 30;

    std::cerr << "WARNING: Unknown bytes unit '" << unit << "'" << std::endl;

    return 0;
}
*/

/**
 \ingroup ssl_crtd
 * Parse uninterrapted string of bytes value. It looks like "4MB".
 */
 /* removed by AG, not referenced anymore
static bool parseBytesOptionValue(size_t * bptr, char const * value)
{
    // Find number from string beginning.
    char const * number_begin = value;
    char const * number_end = value;

    while ((*number_end >= '0' && *number_end <= '9')) {
        ++number_end;
    }

    std::string number(number_begin, number_end - number_begin);
    std::istringstream in(number);
    int d = 0;
    if (!(in >> d))
        return false;

    int m;
    if ((m = parseBytesUnits(number_end)) == 0) {
        return false;
    }

    *bptr = static_cast<size_t>(m * d);
    if (static_cast<long>(*bptr * 2) != m * d * 2)
        return false;

    return true;
}
*/

/**
 \ingroup ssl_crtd
 * Print help using response code.
 */
static void usage()
{
    std::string example_host_name = "host.dom";
    std::string request_string = Ssl::CrtdMessage::param_host + "=" + example_host_name;
    std::stringstream request_string_size_stream;
    request_string_size_stream << request_string.length();
    std::string help_string =
        "usage: ssl_crtd -hv -s ssl_storage_path -M storage_max_size\n"
        "\t-h                   Help\n"
        "\t-v                   Version\n"
        "\t-s ssl_storage_path  Path to specific disk storage of ssl server\n"
        "\t                     certificates.\n"
        "\t-M storage_max_size  max size of ssl certificates storage.\n"
        "\t-b fs_block_size     File system block size in bytes. Need for processing\n"
        "\t                     natural size of certificate on disk. Default value is\n"
        "\t                     2048 bytes.\n"
        "\n"
        "After running write requests in the next format:\n"
        "<request code><whitespace><body_len><whitespace><body>\n"
        "There are two kind of request now:\n"
        + Ssl::CrtdMessage::code_new_certificate + " " + request_string_size_stream.str() + " " + request_string + "\n" +
        "\tCreate new private key and selfsigned certificate for \"host.dom\".\n"
        + Ssl::CrtdMessage::code_new_certificate + " xxx " + request_string + "\n" +
        "-----BEGIN CERTIFICATE-----\n"
        "...\n"
        "-----END CERTIFICATE-----\n"
        "-----BEGIN RSA PRIVATE KEY-----\n"
        "...\n"
        "-----END RSA PRIVATE KEY-----\n"
        "\tCreate new private key and certificate request for \"host.dom\"\n"
        "\tSign new request by received certificate and private key.\n"
        "usage: ssl_crtd -c -s ssl_store_path\n"
        "\t-c                   Init ssl db directories and exit.\n";
    std::cerr << help_string << std::endl;
}

/// Process new request message.
/// Modified by AG to return our known SSL certificate instead of generating a new one 
static bool processNewRequest(Ssl::CrtdMessage & request_message)
{    
    try {

        kantar::log->writeLine("************ Processing request ************");

        Ssl::CertificateProperties certProperties;
        std::string error;
        if (!request_message.parseRequest(certProperties, error))
            throw std::runtime_error("Error while parsing the crtd request: " + error);

        kantar::log->writeLine(std::string("\trequested properties: ") + certProperties.commonName);

        const std::string &cert_subject = certProperties.dbKey();
        kantar::log->writeLine(std::string("\tkey: ") + cert_subject);

        kantar::sendCertificateForDomain(certProperties.commonName);
    }
    catch (std::exception& exc)
    {
        kantar::log->writeLine(std::string("Error in processNewRequest: ") + exc.what());
        throw;
    }

    return true;
}

bool isEnableLoggingMarkerFilePresent()
{
    return access(kantar::paths::kEnableLoggingMarkerFilePath.c_str(), F_OK) != -1;
}

/**
 \ingroup ssl_crtd
 * This is the external ssl_crtd process.
 */
int main(int argc, char *argv[])
{        
    // set up logging
    if (isEnableLoggingMarkerFilePresent())
        kantar::log = new kantar::FileLog();
    else 
        kantar::log = new kantar::NullLog();

    kantar::log->writeLine(std::string("started ") + kantar::toString(kantar::gBuildNumber) + " running as user " + cuserid(nullptr));    

    // load certs and config
    kantar::loadCertificateConfig();

    try {        
        int8_t c;                
        // process options.
        while ((c = getopt(argc, argv, "tdcghvs:M:b:n:")) != -1) {
            switch (c) {
            case 'd':
                debug_enabled = 1;
                break;
            case 'b':
                // removed by AG: we're not using database
                // if (!parseBytesOptionValue(&fs_block_size, optarg)) {
                //     throw std::runtime_error("Error when parsing -b options value");
                // }
                break;
            case 's':
                // removed by AG: we're not using database
                //db_path = optarg;
                break;
            case 'M':
                // removed by AG: we're not using database
                //if (!parseBytesOptionValue(&max_db_size, optarg)) {
                //  throw std::runtime_error("Error when parsing -M options value");
                //}
                break;
            case 'v':
                std::cout << "ssl_crtd version " << VERSION << " build " << kantar::gBuildNumber << std::endl;
                exit(0);
                break;
            case 'c':
                // AG: not using db for cert storage anymore
                //create_new_db = true;
                break;
            case 'h':
                usage();
                exit(0);
            case 't':
            {
                //std::string domain("159.122.243.166");
                std::string domain("77.88.106.124");
                std::cout << "Running test for " << domain << "...\n";
                kantar::sendCertificateForDomain(domain.c_str());
                exit(0);
            }
            default:
                exit(0);
            }
        }        

       
        // Initialize SSL subsystem
        SSL_load_error_strings();
        SSLeay_add_ssl_algorithms();
        

        // process request.
        for (;;) {
            char request[HELPER_INPUT_BUFFER];
            Ssl::CrtdMessage request_message(Ssl::CrtdMessage::REQUEST);
            Ssl::CrtdMessage::ParseResult parse_result = Ssl::CrtdMessage::INCOMPLETE;

            while (parse_result == Ssl::CrtdMessage::INCOMPLETE) {
                if (fgets(request, HELPER_INPUT_BUFFER, stdin) == NULL)
                    return 1;
                size_t gcount = strlen(request);
                parse_result = request_message.parse(request, gcount);
            }

            if (parse_result == Ssl::CrtdMessage::ERROR) {
                throw std::runtime_error("Cannot parse request message.");
            } else if (request_message.getCode() == Ssl::CrtdMessage::code_new_certificate) {
                processNewRequest(request_message);
            } else {
                throw std::runtime_error("Unknown request code: \"" + request_message.getCode() + "\".");
            }
            std::cout.flush();
        }
    } catch (std::runtime_error & error) {
        std::stringstream format;
        format << "ERROR: " << argv[0] << ": " << error.what();
        std::cerr << format.str() << std::endl;
        kantar::log->writeLine(format.str());
        return 0;
    }
    return 0;
}

